from typing import Iterable, Iterator, Mapping

class Dictionary:
    def __getitem__(self, key: str) -> str: ...
    def __setitem__(self, key: str, value: str) -> None: ...
    def __delitem__(self, key: str) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[str]: ...
    def __repr__(self) -> str: ...
    def keys(self) -> Iterable[str]: ...
    def copy(self) -> Dictionary: ...
    def pop(self, key: str) -> str: ...
    def update(
        self,
        other: Mapping[str, str] | Iterable[tuple[str, str]] = (),
        /,
        **kwds: str,
    ) -> None: ...
